from sampo.structurator.base import graph_restructuring, STAGE_SEP
from sampo.structurator.graph_insertion import graph_in_graph_insertion
from sampo.structurator.light_modifications import work_graph_ids_simplification
